@@ -1,13 +1,13 @@
 /* 
- * Copyright (c) 2008-2010 Wind River Systems; see
+ * Copyright (c) 2008-2010, 2012 Wind River Systems; see
  * guts/COPYRIGHT for information.
  *
  * static int
  * wrap_remove(const char *path) {
  *	int rc = -1;
  */
-	struct stat buf;
-	if (real_lstat(path, &buf) == -1) {
+	PSEUDO_STATBUF buf;
+	if (base_lstat(path, &buf) == -1) {
 		errno = ENOENT;
 		return -1;
 	}
