@@ -23,6 +23,7 @@
 #include <sys/stat.h>
 #include <sys/xattr.h>
 #include <time.h>
+#include <sys/time.h>
 #include <unistd.h>
 
 #include <sqlite3.h>
@@ -65,6 +66,18 @@ static sqlite3 *log_db = 0;
  */
 typedef struct { char *fmt; int arg; } id_row;
 
+/* So the migrations were getting to be large and complicated. Phase
+ * one of getting rid of that is to merge their effects in and set up
+ * a dummy migration to indicate that this was done.
+ */
+static id_row files_default_migrations[] = {
+	{ "%d, -1, 'N/A (initial setup)'", 8 },
+	{ NULL },
+};
+static id_row logs_default_migrations[] = {
+	{ "%d, -1, 'N/A (initial setup)'", 7 },
+	{ NULL },
+};
 /* This seemed like a really good idea at the time.  The idea is that these
  * structures let me write semi-abstract code to "create a database" without
  * duplicating as much of the code.
@@ -72,7 +85,7 @@ typedef struct { char *fmt; int arg; } id_row;
 static struct sql_table {
 	char *name;
 	char *sql;
-	char **names;
+	char *names;
 	id_row *values;
 } file_tables[] = {
 	{ "files",
@@ -83,16 +96,25 @@ static struct sql_table {
 	    "uid INTEGER, "
 	    "gid INTEGER, "
 	    "mode INTEGER, "
-	    "rdev INTEGER",
+	    "rdev INTEGER, "
+	    "deleting INTEGER",
 	  NULL,
 	  NULL },
 	{ "xattrs",
 	  "id INTEGER PRIMARY KEY, "
-	    "file_id INTEGER REFERENCES files(id) ON DELETE CASCADE, "
+	    "dev INTEGER, "
+	    "ino INTEGER, "
 	    "name VARCHAR, "
 	    "value VARCHAR",
 	  NULL,
 	  NULL },
+	{ "migrations",
+		"id INTEGER PRIMARY KEY, "
+		"version INTEGER, "
+		"stamp INTEGER, "
+		"sql VARCHAR",
+	  "version, stamp, sql",
+	  files_default_migrations },
 	{ NULL, NULL, NULL, NULL },
 }, log_tables[] = {
 	{ "logs",
@@ -107,9 +129,23 @@ static struct sql_table {
 	    "path VARCHAR, "
 	    "result INTEGER, "
 	    "severity INTEGER, "
-	    "text VARCHAR ",
+	    "text VARCHAR, "
+	    "tag VARCHAR, "
+	    "uid INTEGER, "
+	    "gid INTEGER, "
+	    "access INTEGER, "
+	    "program VARCHAR, "
+	    "type INTEGER, "
+	    ,
 	  NULL,
 	  NULL },
+	{ "migrations",
+		"id INTEGER PRIMARY KEY, "
+		"version INTEGER, "
+		"stamp INTEGER, "
+		"sql VARCHAR",
+	  "version, stamp, sql",
+	  logs_default_migrations },
 	{ NULL, NULL, NULL, NULL },
 };
 
@@ -119,12 +155,14 @@ static struct sql_index {
 	char *table;
 	char *keys;
 } file_indexes[] = {
-/*	{ "files__path", "files", "path" }, */
 	{ "files__path_dev_ino", "files", "path, dev, ino" },
 	{ "files__dev_ino", "files", "dev, ino" },
-	{ "xattrs__file", "xattrs", "file_id" },
+	{ "migration__version", "migrations", "version" },
+	{ "xattrs_dev_ino", "xattrs", "dev, ino" },
+	{ "xattrs_dev_ino_name", "xattrs", "dev, ino, name" },
 	{ NULL, NULL, NULL },
 }, log_indexes[] = {
+	{ "migration__version", "migrations", "version" },
 	{ NULL, NULL, NULL },
 };
 
@@ -193,6 +231,47 @@ static struct sql_migration {
 	{ create_migration_table },
 	{ index_migration_table },
 	{ "ALTER TABLE files ADD deleting INTEGER;" },
+	/* oops. I made a design mistake, and the table is in the wrong
+	 * format. But sqlite doesn't support modifying constraints on an
+	 * existing table, or dropping columns. Soooooo.
+	 *
+	 * Also, you can't have a foreign key relationship to data which
+	 * aren't actually unique, but the entire problem is that dev/ino
+	 * are not unique.
+	 *
+	 * Note that the name entry is not really all that useful; it's mostly
+	 * there to help with debugging, now.
+	 */
+	{
+	  "CREATE TABLE xattrs_new ("
+	      "id INTEGER PRIMARY KEY, "
+	      "dev INTEGER, "
+	      "ino INTEGER, "
+	      "name VARCHAR, "
+	      "value VARCHAR"
+	    ");",
+	},
+	{ 
+	  "INSERT INTO xattrs_new (dev, ino, name, value) "
+	    "SELECT "
+	      "(SELECT dev FROM files WHERE id = file_id), "
+	      "(SELECT ino FROM files WHERE id = file_id), "
+	      "name, "
+	      "value "
+	    "FROM xattrs;"
+	},
+	{
+	  "DROP TABLE xattrs;"
+	},
+	{
+	  "ALTER TABLE xattrs_new RENAME TO xattrs;",
+	},
+	{
+	  "CREATE INDEX xattrs__dev_ino ON xattrs (dev, ino);",
+	},
+	{
+	  "CREATE UNIQUE INDEX xattrs__dev_ino_name ON xattrs (dev, ino, name);",
+	},
 	{ NULL },
 }, log_migrations[] = {
 	{ create_migration_table },
@@ -399,8 +478,9 @@ make_tables(sqlite3 *db,
 				snprintf(buffer, sizeof(buffer), sql_tables[i].values[j].fmt, sql_tables[i].values[j].arg);
 				sql = sqlite3_mprintf("INSERT INTO %s ( %s ) VALUES ( %s );",
 					sql_tables[i].name,
-					*sql_tables[i].names,
+					sql_tables[i].names,
 					buffer);
+				pseudo_debug(PDBGF_DB, "data setup: %s\n", sql);
 				rc = sqlite3_exec(db, sql, NULL, NULL, &errmsg);
 				sqlite3_free(sql);
 				if (rc) {
@@ -426,51 +506,46 @@ make_tables(sqlite3 *db,
 			}
 		}
 	}
-	/* now, see about migrations */
-	found = 0;
-	for (j = 1; j <= rows; ++j) {
-		if (!strcmp(existing[j], "migrations")) {
-			found = 1;
-			break;
-		}
+
+	/* the migrations table is assumed to exist, because we now
+	 * create it automatically; in a previous implementation, it was
+	 * only created by the first migration.
+	 */
+	sql = "SELECT MAX(version) FROM migrations;";
+	rc = sqlite3_prepare_v2(db, sql, strlen(sql), &stmt, NULL);
+	if (rc) {
+		dberr(db, "couldn't examine migrations table");
+		return 1;
 	}
-	if (found) {
-		sql = "SELECT MAX(version) FROM migrations;";
-		rc = sqlite3_prepare_v2(db, sql, strlen(sql), &stmt, NULL);
-		if (rc) {
-			dberr(db, "couldn't examine migrations table");
-			return 1;
-		}
+	rc = sqlite3_step(stmt);
+	if (rc == SQLITE_ROW) {
+		version = (unsigned long) sqlite3_column_int64(stmt, 0);
 		rc = sqlite3_step(stmt);
-		if (rc == SQLITE_ROW) {
-			version = (unsigned long) sqlite3_column_int64(stmt, 0);
-			rc = sqlite3_step(stmt);
-		} else {
-			version = -1;
-		}
-		if (rc != SQLITE_DONE) {
-			dberr(db, "not done after the single row we expected?", rc);
-			return 1;
-		}
-		pseudo_debug(PDBGF_DB, "existing database version: %d\n", version);
-		rc = sqlite3_finalize(stmt);
-		if (rc) {
-			dberr(db, "couldn't finalize version check");
-			return 1;
-		}
 	} else {
-		pseudo_debug(PDBGF_DB, "no existing database version\n");
 		version = -1;
 	}
+	if (rc != SQLITE_DONE) {
+		dberr(db, "not done after the single row we expected?", rc);
+		return 1;
+	}
+	pseudo_debug(PDBGF_DB, "existing database version: %d\n", version);
+	rc = sqlite3_finalize(stmt);
+	if (rc) {
+		dberr(db, "couldn't finalize version check");
+		return 1;
+	}
 	for (m = sql_migrations; m->sql; ++m)
 		;
-	available_migrations = m - sql_migrations;
+	/* m points to the null, so available migrations is one lower */
+	available_migrations = m - sql_migrations - 1;
 	/* I am pretty sure this can never happen. */
 	if (version < -1)
 		version = -1;
 	/* I hope this can never happen. */
-	if (version >= available_migrations)
-		version = available_migrations - 1;
+	pseudo_debug(PDBGF_DB, "version %d, available_migrations %d\n",
+		version, available_migrations);
+	if (version > available_migrations)
+		version = available_migrations;
 	for (m = sql_migrations + (version + 1); m->sql; ++m) {
 		int migration = (m - sql_migrations);
 		pseudo_debug(PDBGF_DB, "considering migration %d\n", migration);
@@ -537,17 +612,36 @@ make_tables(sqlite3 *db,
 /* registered with atexit */
 static void
 cleanup_db(void) {
-	pseudo_debug(PDBGF_SERVER, "server exiting\n");
+	char datebuf[64];
+	struct tm stamp_tm;
+	struct timeval stamp;
+
+	gettimeofday(&stamp, NULL);
+	localtime_r(&stamp.tv_sec, &stamp_tm);
+	strftime(datebuf, 64, "%H:%M:%S", &stamp_tm);
+
+	pseudo_diag("db cleanup for server shutdown, %s.%03d\n",
+		datebuf, (int) (stamp.tv_usec / 1000));
 #ifdef USE_MEMORY_DB
         if (real_file_db) {
                 pdb_backup();
                 sqlite3_close(real_file_db);
 	}
+	gettimeofday(&stamp, NULL);
+	localtime_r(&stamp.tv_sec, &stamp_tm);
+	strftime(datebuf, 64, "%H:%M:%S", &stamp_tm);
+	pseudo_diag("memory-to-file backup complete, %s.%03d.\n",
+		datebuf, (int) (stamp.tv_usec / 1000));
 #endif
 	if (file_db)
 		sqlite3_close(file_db);
 	if (log_db)
 		sqlite3_close(log_db);
+	gettimeofday(&stamp, NULL);
+	localtime_r(&stamp.tv_sec, &stamp_tm);
+	strftime(datebuf, 64, "%H:%M:%S", &stamp_tm);
+	pseudo_diag("db cleanup finished, %s.%03d\n",
+		datebuf, (int) (stamp.tv_usec / 1000));
 }
 
 /* This function has been rewritten and I no longer hate it.  I just feel
@@ -573,6 +667,7 @@ get_db(struct database_info *dbinfo) {
 	if (*(dbinfo->db))
 		return 0;
 
+	pseudo_debug(PDBGF_DB, "get_db(%s)\n", dbinfo->pathname);
 	dbfile = pseudo_localstatedir_path(dbinfo->pathname);
 #ifdef USE_MEMORY_DB
         if (!strcmp(dbinfo->pathname, ":memory:")) {
@@ -1365,11 +1460,158 @@ log_entry_free(log_entry *e) {
 
 /* Now for the actual file handling code! */
 
+/* nuke any unlinked xattrs. Only used from the rare case where we try
+ * to nuke a whole bunch of "deleting" files.
+ */
+
+static void
+pdb_clear_unused_xattrs(void) {
+	static sqlite3_stmt *delete;
+	char *delete_sql = "DELETE FROM xattrs WHERE (SELECT COUNT(*) FROM files WHERE dev = xattrs.dev AND ino = xattrs.ino) = 0;";
+	int rc;
+
+	if (!file_db && get_dbs()) {
+		pseudo_diag("%s: database error.\n", __func__);
+		return;
+	}
+	if (!delete) {
+		rc = sqlite3_prepare_v2(file_db, delete_sql, strlen(delete_sql), &delete, NULL);
+		if (rc) {
+			dberr(file_db, "couldn't prepare DELETE statement for unused xattrs");
+			return;
+		}
+	}
+	rc = sqlite3_step(delete);
+	if (rc != SQLITE_DONE) {
+		dberr(file_db, "delete of unused xattrs may have failed");
+	}
+	sqlite3_reset(delete);
+	sqlite3_clear_bindings(delete);
+}
+
+/* we want to delete extended attributes for a device/inode if there
+ * are no longer any links with it. Anything that deletes at least one
+ * link for a given device/inode should then call this.
+ */
+static void
+pdb_clear_xattrs(pseudo_msg_t *msg) {
+	static sqlite3_stmt *delete;
+	char *delete_sql = "DELETE FROM xattrs WHERE dev = ? AND ino = ?;";
+	int rc;
+
+	if (!msg)
+		return;
+	if (!file_db && get_dbs()) {
+		pseudo_diag("%s: database error.\n", __func__);
+		return;
+	}
+	if (!delete) {
+		rc = sqlite3_prepare_v2(file_db, delete_sql, strlen(delete_sql), &delete, NULL);
+		if (rc) {
+			dberr(file_db, "couldn't prepare DELETE statement for unused xattrs");
+			return;
+		}
+	}
+	sqlite3_bind_int(delete, 1, msg->dev);
+	sqlite3_bind_int(delete, 2, msg->ino);
+	rc = sqlite3_step(delete);
+	if (rc != SQLITE_DONE) {
+		dberr(file_db, "delete of unused xattrs may have failed");
+	}
+	sqlite3_reset(delete);
+	sqlite3_clear_bindings(delete);
+}
+
+static void
+pdb_copy_xattrs(pseudo_msg_t *oldmsg, pseudo_msg_t *msg) {
+	static sqlite3_stmt *copy;
+	char *copy_sql = "INSERT INTO xattrs (dev, ino, name, value) "
+	    "SELECT "
+	      "?, "
+	      "?, "
+	      "name, "
+	      "value "
+	    "FROM xattrs WHERE dev = ? AND ino = ?;";
+	int rc;
+
+	if (!oldmsg || !msg)
+		return;
+	if (!file_db && get_dbs()) {
+		pseudo_diag("%s: database error.\n", __func__);
+		return;
+	}
+	if (!copy) {
+		rc = sqlite3_prepare_v2(file_db, copy_sql, strlen(copy_sql), &copy, NULL);
+		if (rc) {
+			dberr(file_db, "couldn't prepare INSERT statement for copying xattrs");
+			return;
+		}
+	}
+	sqlite3_bind_int(copy, 1, msg->dev);
+	sqlite3_bind_int(copy, 2, msg->ino);
+	sqlite3_bind_int(copy, 3, oldmsg->dev);
+	sqlite3_bind_int(copy, 4, oldmsg->ino);
+	rc = sqlite3_step(copy);
+	if (rc != SQLITE_DONE) {
+		dberr(file_db, "copy of xattrs may have failed");
+	}
+	sqlite3_reset(copy);
+	sqlite3_clear_bindings(copy);
+}
+
+static void
+pdb_check_xattrs(pseudo_msg_t *msg) {
+	static sqlite3_stmt *scan;
+	char *scan_sql = "SELECT COUNT(*) FROM files WHERE dev = ? AND ino = ?;";
+	int rc;
+
+	if (!msg)
+		return;
+	if (!file_db && get_dbs()) {
+		pseudo_diag("%s: database error.\n", __func__);
+		return;
+	}
+	if (!scan) {
+		rc = sqlite3_prepare_v2(file_db, scan_sql, strlen(scan_sql), &scan, NULL);
+		if (rc) {
+			dberr(file_db, "couldn't prepare SELECT statement for checking for existing files");
+			return;
+		}
+	}
+	int existing;
+	sqlite3_bind_int(scan, 1, msg->dev);
+	sqlite3_bind_int(scan, 2, msg->ino);
+	rc = sqlite3_step(scan);
+	if (rc == SQLITE_ROW) {
+		existing = (int) sqlite3_column_int64(scan, 0);
+		pseudo_debug(PDBGF_DB | PDBGF_VERBOSE,
+			"scan for existing files: %d\n",
+			existing);
+		rc = sqlite3_step(scan);
+	} else {
+		pseudo_debug(PDBGF_DB, "scan for existing files: no count?\n");
+		existing = 0;
+	}
+	if (rc != SQLITE_DONE) {
+		dberr(file_db, "not done after the single row we expected?", rc);
+	}
+	rc = sqlite3_finalize(scan);
+	if (rc) {
+		dberr(file_db, "couldn't finalize existing file");
+	}
+	sqlite3_reset(scan);
+	sqlite3_clear_bindings(scan);
+
+	/* if there are any matching entries, don't delete xattrs */
+	if (existing == 0)
+		pdb_clear_xattrs(msg);
+}
+
 /* pdb_link_file:  Creates a new file from msg, using the provided path
  * or 'NAMELESS FILE'.
  */
 int
-pdb_link_file(pseudo_msg_t *msg, long long *row) {
+pdb_link_file(pseudo_msg_t *msg) {
 	static sqlite3_stmt *insert;
 	int rc;
 	char *sql = "INSERT INTO files "
@@ -1410,10 +1652,6 @@ pdb_link_file(pseudo_msg_t *msg, long long *row) {
 	if (rc != SQLITE_DONE) {
 		dberr(file_db, "insert may have failed (rc %d)", rc);
 	}
-	/* some users care what the row ID is */
-	if (row) {
-		*row = sqlite3_last_insert_rowid(file_db);
-	}
 	sqlite3_reset(insert);
 	sqlite3_clear_bindings(insert);
 	return rc != SQLITE_DONE;
@@ -1449,6 +1687,7 @@ pdb_unlink_file_dev(pseudo_msg_t *msg) {
 	}
 	sqlite3_reset(sql_delete);
 	sqlite3_clear_bindings(sql_delete);
+	pdb_clear_xattrs(msg);
 	return rc != SQLITE_DONE;
 }
 
@@ -1569,6 +1808,7 @@ pdb_cancel_unlink_file(pseudo_msg_t *msg) {
 	pseudo_debug(PDBGF_DB, "(exact %d) ", exact);
 	sqlite3_reset(mark_file);
 	sqlite3_clear_bindings(mark_file);
+	pdb_check_xattrs(msg);
 	return rc != SQLITE_DONE;
 }
 
@@ -1606,6 +1846,7 @@ pdb_did_unlink_files(int deleting) {
 	pseudo_debug(PDBGF_DB, "(exact %d)\n", exact);
 	sqlite3_reset(delete_exact);
 	sqlite3_clear_bindings(delete_exact);
+	pdb_clear_unused_xattrs();
 	return rc != SQLITE_DONE;
 }
 
@@ -1642,6 +1883,10 @@ pdb_did_unlink_file(char *path, int deleting) {
 	pseudo_debug(PDBGF_DB, "(exact %d)\n", exact);
 	sqlite3_reset(delete_exact);
 	sqlite3_clear_bindings(delete_exact);
+	/* we have to clean everything because we don't know for sure the
+	 * device/inode...
+	 */
+	pdb_clear_unused_xattrs();
 	return rc != SQLITE_DONE;
 }
 
@@ -1681,6 +1926,7 @@ pdb_unlink_file(pseudo_msg_t *msg) {
 	pseudo_debug(PDBGF_DB, "(exact %d) ", exact);
 	sqlite3_reset(delete_exact);
 	sqlite3_clear_bindings(delete_exact);
+	pdb_check_xattrs(msg);
 	return rc != SQLITE_DONE;
 }
 
@@ -1729,6 +1975,8 @@ pdb_unlink_contents(pseudo_msg_t *msg) {
 	pseudo_debug(PDBGF_DB, "(sub %d) ", sub);
 	sqlite3_reset(delete_sub);
 	sqlite3_clear_bindings(delete_sub);
+	/* we have no idea how many things might have been affected, so. */
+	pdb_clear_unused_xattrs();
 	return rc != SQLITE_DONE;
 }
 
@@ -1813,9 +2061,12 @@ pdb_rename_file(const char *oldpath, pseudo_msg_t *msg) {
  */
 int
 pdb_renumber_all(dev_t from, dev_t to) {
-	static sqlite3_stmt *update;
+	static sqlite3_stmt *files_update, *xattrs_update;
 	int rc;
-	char *sql = "UPDATE files "
+	char *files_sql = "UPDATE files "
+		    " SET dev = ? "
+		    " WHERE dev = ?;";
+	char *xattrs_sql = "UPDATE xattrs "
 		    " SET dev = ? "
 		    " WHERE dev = ?;";
 
@@ -1823,44 +2074,75 @@ pdb_renumber_all(dev_t from, dev_t to) {
 		pseudo_diag("%s: database error.\n", __func__);
 		return 0;
 	}
-	if (!update) {
-		rc = sqlite3_prepare_v2(file_db, sql, strlen(sql), &update, NULL);
+	if (!files_update) {
+		rc = sqlite3_prepare_v2(file_db, files_sql, strlen(files_sql), &files_update, NULL);
 		if (rc) {
 			dberr(file_db, "couldn't prepare UPDATE statement");
 			return 1;
 		}
 	}
-	rc = sqlite3_bind_int(update, 1, to);
+	if (!xattrs_update) {
+		rc = sqlite3_prepare_v2(file_db, xattrs_sql, strlen(xattrs_sql), &xattrs_update, NULL);
+		if (rc) {
+			dberr(file_db, "couldn't prepare UPDATE statement");
+			return 1;
+		}
+	}
+	rc = sqlite3_bind_int(files_update, 1, to);
 	if (rc) {
 		dberr(file_db, "error binding device numbers to update");
 	}
-	rc = sqlite3_bind_int(update, 2, from);
+	rc = sqlite3_bind_int(files_update, 2, from);
+	if (rc) {
+		dberr(file_db, "error binding device numbers to update");
+	}
+	rc = sqlite3_bind_int(xattrs_update, 1, to);
+	if (rc) {
+		dberr(file_db, "error binding device numbers to update");
+	}
+	rc = sqlite3_bind_int(xattrs_update, 2, from);
 	if (rc) {
 		dberr(file_db, "error binding device numbers to update");
 	}
 
 	file_db_dirty = 1;
-	rc = sqlite3_step(update);
+	pseudo_debug(PDBGF_DB, "updating device dev %llu to %llu\n",
+		(unsigned long long) from, (unsigned long long) to);
+	rc = sqlite3_step(files_update);
 	if (rc != SQLITE_DONE) {
 		dberr(file_db, "update may have failed: rc %d", rc);
 	}
-	sqlite3_reset(update);
-	sqlite3_clear_bindings(update);
-	pseudo_debug(PDBGF_DB, "updating device dev %llu to %llu\n",
-		(unsigned long long) from, (unsigned long long) to);
+	sqlite3_reset(files_update);
+	sqlite3_clear_bindings(files_update);
+	rc = sqlite3_step(xattrs_update);
+	if (rc != SQLITE_DONE) {
+		dberr(file_db, "update may have failed: rc %d", rc);
+	}
+	sqlite3_reset(xattrs_update);
+	sqlite3_clear_bindings(xattrs_update);
 	return rc != SQLITE_DONE;
 }
 
 /* change dev/inode for a given path -- used only by RENAME for now.
+ * So, if you rename something with xattrs, we want to keep the xattrs.
+ * But. If you renamed one of multiple hard links, we may end up wanting
+ * to keep the xattrs on both the old name and the new one. But if there's
+ * only one link, we want to drop the xattrs from the old one. Eww.
  */
 int
 pdb_update_inode(pseudo_msg_t *msg) {
 	static sqlite3_stmt *update;
 	int rc;
+	static pseudo_msg_t *oldmsg;
+	int found_existing;
+
+	if (!oldmsg) {
+		oldmsg = malloc(pseudo_path_max());
+	}
+
 	char *sql = "UPDATE files "
 		    " SET dev = ?, ino = ? "
 		    " WHERE path = ?;";
-
 	if (!file_db && get_dbs()) {
 		pseudo_diag("%s: database error.\n", __func__);
 		return 0;
@@ -1879,6 +2161,12 @@ pdb_update_inode(pseudo_msg_t *msg) {
 		pseudo_diag("Can't update the inode of a file without its path.\n");
 		return 1;
 	}
+	memcpy(oldmsg, msg, sizeof(msg) + msg->pathlen);
+	found_existing = !pdb_find_file_path(oldmsg);
+	if (found_existing) {
+		/* we have an existing file entry */
+		pdb_copy_xattrs(oldmsg, msg);
+	}
 	sqlite3_bind_int(update, 1, msg->dev);
 	sqlite3_bind_int64(update, 2, msg->ino);
 	rc = sqlite3_bind_text(update, 3, msg->path, -1, SQLITE_STATIC);
@@ -1897,6 +2185,10 @@ pdb_update_inode(pseudo_msg_t *msg) {
 	}
 	sqlite3_reset(update);
 	sqlite3_clear_bindings(update);
+	if (found_existing) {
+		/* possibly delete old message's xattrs */
+		pdb_check_xattrs(oldmsg);
+	}
 	pseudo_debug(PDBGF_DB, "updating path %s to dev %llu, ino %llu\n",
 		msg->path,
 		(unsigned long long) msg->dev, (unsigned long long) msg->ino);
@@ -1950,7 +2242,7 @@ pdb_update_file(pseudo_msg_t *msg) {
 
 /* find file using both path AND dev/inode as key */
 int
-pdb_find_file_exact(pseudo_msg_t *msg, long long *row) {
+pdb_find_file_exact(pseudo_msg_t *msg) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *sql = "SELECT * FROM files WHERE path = ? AND dev = ? AND ino = ?;";
@@ -1978,9 +2270,6 @@ pdb_find_file_exact(pseudo_msg_t *msg, long long *row) {
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
-		if (row) {
-			*row = sqlite3_column_int64(select, 0);
-		}
 		msg->uid = (unsigned long) sqlite3_column_int64(select, 4);
 		msg->gid = (unsigned long) sqlite3_column_int64(select, 5);
 		msg->mode = (unsigned long) sqlite3_column_int64(select, 6);
@@ -1989,7 +2278,7 @@ pdb_find_file_exact(pseudo_msg_t *msg, long long *row) {
 		rc = 0;
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(PDBGF_DB, "find_exact: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB | PDBGF_VERBOSE, "find_exact: sqlite_done on first row\n");
 		rc = 1;
 		break;
 	default:
@@ -2004,7 +2293,7 @@ pdb_find_file_exact(pseudo_msg_t *msg, long long *row) {
 
 /* find file using path as a key */
 int
-pdb_find_file_path(pseudo_msg_t *msg, long long *row) {
+pdb_find_file_path(pseudo_msg_t *msg) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *sql = "SELECT * FROM files WHERE path = ?;";
@@ -2035,9 +2324,6 @@ pdb_find_file_path(pseudo_msg_t *msg, long long *row) {
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
-		if (row) {
-			*row = sqlite3_column_int64(select, 0);
-		}
 		msg->dev = sqlite3_column_int64(select, 2);
 		msg->ino = sqlite3_column_int64(select, 3);
 		msg->uid = sqlite3_column_int64(select, 4);
@@ -2048,7 +2334,7 @@ pdb_find_file_path(pseudo_msg_t *msg, long long *row) {
 		rc = 0;
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(PDBGF_DB, "find_path: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB | PDBGF_VERBOSE, "find_path: sqlite_done on first row\n");
 		rc = 1;
 		break;
 	default:
@@ -2098,7 +2384,7 @@ pdb_get_file_path(pseudo_msg_t *msg) {
 		}
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(PDBGF_DB, "find_dev: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB | PDBGF_VERBOSE, "find_dev: sqlite_done on first row\n");
 		response = 0;
 		break;
 	default:
@@ -2113,7 +2399,7 @@ pdb_get_file_path(pseudo_msg_t *msg) {
 
 /* find file using dev/inode as key */
 int
-pdb_find_file_dev(pseudo_msg_t *msg, long long *row, char **path) {
+pdb_find_file_dev(pseudo_msg_t *msg, char **path) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *sql = "SELECT * FROM files WHERE dev = ? AND ino = ?;";
@@ -2137,9 +2423,6 @@ pdb_find_file_dev(pseudo_msg_t *msg, long long *row, char **path) {
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
-		if (row) {
-			*row = sqlite3_column_int64(select, 0);
-		}
 		msg->uid = (unsigned long) sqlite3_column_int64(select, 4);
 		msg->gid = (unsigned long) sqlite3_column_int64(select, 5);
 		msg->mode = (unsigned long) sqlite3_column_int64(select, 6);
@@ -2154,7 +2437,7 @@ pdb_find_file_dev(pseudo_msg_t *msg, long long *row, char **path) {
 		rc = 0;
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(PDBGF_DB, "find_dev: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB | PDBGF_VERBOSE, "find_dev: sqlite_done on first row\n");
 		rc = 1;
 		break;
 	default:
@@ -2168,12 +2451,12 @@ pdb_find_file_dev(pseudo_msg_t *msg, long long *row, char **path) {
 }
 
 int
-pdb_get_xattr(long long file_id, char **value, size_t *len) {
+pdb_get_xattr(pseudo_msg_t *msg, char **value, size_t *len) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *response;
 	size_t length;
-	char *sql = "SELECT value FROM xattrs WHERE file_id = ? AND name = ?;";
+	char *sql = "SELECT value FROM xattrs WHERE dev  = ? AND ino = ? AND name = ?;";
 
 	if (!file_db && get_dbs()) {
 		pseudo_diag("%s: database error.\n", __func__);
@@ -2186,9 +2469,10 @@ pdb_get_xattr(long long file_id, char **value, size_t *len) {
 			return 1;
 		}
 	}
-	pseudo_debug(PDBGF_XATTR, "requested xattr named '%s' for file %lld\n", *value, file_id);
-	sqlite3_bind_int(select, 1, file_id);
-	rc = sqlite3_bind_text(select, 2, *value, -1, SQLITE_STATIC);
+	pseudo_debug(PDBGF_XATTR, "requested xattr named '%s' for ino %lld\n", *value, (long long) msg->ino);
+	sqlite3_bind_int(select, 1, msg->dev);
+	sqlite3_bind_int(select, 2, msg->ino);
+	rc = sqlite3_bind_text(select, 3, *value, -1, SQLITE_STATIC);
 	if (rc) {
 		dberr(file_db, "couldn't bind xattr name to SELECT.");
 		return 1;
@@ -2215,7 +2499,7 @@ pdb_get_xattr(long long file_id, char **value, size_t *len) {
 		}
 		break;
 	case SQLITE_DONE:
-		pseudo_debug(PDBGF_DB, "find_exact: sqlite_done on first row\n");
+		pseudo_debug(PDBGF_DB | PDBGF_VERBOSE, "find_exact: sqlite_done on first row\n");
 		rc = 1;
 		break;
 	default:
@@ -2229,22 +2513,13 @@ pdb_get_xattr(long long file_id, char **value, size_t *len) {
 }
 
 int
-pdb_list_xattr(long long file_id, char **value, size_t *len) {
+pdb_list_xattr(pseudo_msg_t *msg, char **value, size_t *len) {
 	static sqlite3_stmt *select;
 	size_t allocated = 0;
 	size_t used = 0;
 	char *buffer = 0;
 	int rc;
-	char *sql = "SELECT name FROM xattrs WHERE file_id = ? ORDER BY name;";
-
-	/* if we don't have a record of the file, it must not have
-	 * any extended attributes...
-	 */
-	if (file_id == -1) {
-		*value = NULL;
-		*len = 0;
-		return 0;
-	}
+	char *sql = "SELECT name FROM xattrs WHERE dev = ? AND ino = ? ORDER BY name;";
 
 	if (!file_db && get_dbs()) {
 		pseudo_diag("%s: database error.\n", __func__);
@@ -2257,7 +2532,8 @@ pdb_list_xattr(long long file_id, char **value, size_t *len) {
 			return 1;
 		}
 	}
-	sqlite3_bind_int(select, 1, file_id);
+	sqlite3_bind_int(select, 1, msg->dev);
+	sqlite3_bind_int(select, 2, msg->ino);
 	do {
 		rc = sqlite3_step(select);
 		if (rc == SQLITE_ROW) {
@@ -2294,10 +2570,10 @@ pdb_list_xattr(long long file_id, char **value, size_t *len) {
 }
 
 int
-pdb_remove_xattr(long long file_id, char *value, size_t len) {
+pdb_remove_xattr(pseudo_msg_t *msg, char *value, size_t len) {
 	static sqlite3_stmt *delete;
 	int rc;
-	char *sql = "DELETE FROM xattrs WHERE file_id = ? AND name = ?;";
+	char *sql = "DELETE FROM xattrs WHERE dev = ? AND ino = ? AND name = ?;";
 
 	if (!file_db && get_dbs()) {
 		pseudo_diag("%s: database error.\n", __func__);
@@ -2310,8 +2586,9 @@ pdb_remove_xattr(long long file_id, char *value, size_t len) {
 			return 1;
 		}
 	}
-	sqlite3_bind_int(delete, 1, file_id);
-	rc = sqlite3_bind_text(delete, 2, value, len, SQLITE_STATIC);
+	sqlite3_bind_int(delete, 1, msg->dev);
+	sqlite3_bind_int(delete, 2, msg->ino);
+	rc = sqlite3_bind_text(delete, 3, value, len, SQLITE_STATIC);
 	if (rc) {
 		dberr(file_db, "couldn't bind xattr name to DELETE.");
 		return 1;
@@ -2327,14 +2604,14 @@ pdb_remove_xattr(long long file_id, char *value, size_t len) {
 }
 
 int
-pdb_set_xattr(long long file_id, char *value, size_t len, int flags) {
+pdb_set_xattr(pseudo_msg_t *msg, char *value, size_t len, int flags) {
 	static sqlite3_stmt *select, *update, *insert;
 	int rc;
 	long long existing_row = -1;
-	char *select_sql = "SELECT id FROM xattrs WHERE file_id = ? AND name = ?;";
+	char *select_sql = "SELECT id FROM xattrs WHERE dev = ? AND ino = ? AND name = ?;";
 	char *insert_sql = "INSERT INTO xattrs "
-		    " ( file_id, name, value ) "
-		    " VALUES (?, ?, ?);";
+		    " ( dev, ino, name, value ) "
+		    " VALUES (?, ?, ?, ?);";
 	char *update_sql = "UPDATE xattrs SET value = ? WHERE id = ?;";
 	char *vname = value;
 	size_t vlen;
@@ -2350,8 +2627,9 @@ pdb_set_xattr(long long file_id, char *value, size_t len, int flags) {
 			return 1;
 		}
 	}
-	sqlite3_bind_int(select, 1, file_id);
-	rc = sqlite3_bind_text(select, 2, value, -1, SQLITE_STATIC);
+	sqlite3_bind_int(select, 1, msg->dev);
+	sqlite3_bind_int(select, 2, msg->ino);
+	rc = sqlite3_bind_text(select, 3, value, -1, SQLITE_STATIC);
 	if (rc) {
 		dberr(file_db, "couldn't bind xattr name to SELECT.");
 		return 1;
@@ -2384,8 +2662,8 @@ pdb_set_xattr(long long file_id, char *value, size_t len, int flags) {
 	vlen = strlen(value);
 	len = len - (vlen + 1);
 	value = value + vlen + 1;
-	pseudo_debug(PDBGF_XATTR, "trying to set a value for %lld: name is '%s' [%d/%d bytes], value is '%s' [%d bytes]. Existing row %lld.\n",
-		file_id, vname, (int) vlen, (int) (len + vlen + 1), value, (int) len, existing_row);
+	pseudo_debug(PDBGF_XATTR, "trying to set a value for ino %lld: name is '%s' [%d/%d bytes], value is '%s' [%d bytes]. Existing row %lld.\n",
+		(long long) msg->ino, vname, (int) vlen, (int) (len + vlen + 1), value, (int) len, existing_row);
 	if (existing_row != -1) {
 		/* update */
 		if (!update) {
@@ -2418,14 +2696,14 @@ pdb_set_xattr(long long file_id, char *value, size_t len, int flags) {
 				return 1;
 			}
 		}
-		pseudo_debug(PDBGF_XATTR, "insert should be getting ID %lld\n", file_id);
-		sqlite3_bind_int64(insert, 1, file_id);
-		rc = sqlite3_bind_text(insert, 2, vname, -1, SQLITE_STATIC);
+		sqlite3_bind_int64(insert, 1, msg->dev);
+		sqlite3_bind_int64(insert, 2, msg->ino);
+		rc = sqlite3_bind_text(insert, 3, vname, -1, SQLITE_STATIC);
 		if (rc) {
 			dberr(file_db, "couldn't bind xattr name to INSERT statement");
 			return 1;
 		}
-		rc = sqlite3_bind_text(insert, 3, value, len, SQLITE_STATIC);
+		rc = sqlite3_bind_text(insert, 4, value, len, SQLITE_STATIC);
 		if (rc) {
 			dberr(file_db, "couldn't bind xattr value to INSERT statement");
 			return 1;
@@ -2446,7 +2724,7 @@ pdb_set_xattr(long long file_id, char *value, size_t len, int flags) {
  * in for NFS usage.
  */
 int
-pdb_find_file_ino(pseudo_msg_t *msg, long long *row) {
+pdb_find_file_ino(pseudo_msg_t *msg) {
 	static sqlite3_stmt *select;
 	int rc;
 	char *sql = "SELECT * FROM files WHERE ino = ?;";
@@ -2469,9 +2747,6 @@ pdb_find_file_ino(pseudo_msg_t *msg, long long *row) {
 	rc = sqlite3_step(select);
 	switch (rc) {
 	case SQLITE_ROW:
-		if (row) {
-			*row = sqlite3_column_int64(select, 0);
-		}
 		msg->dev = (unsigned long) sqlite3_column_int64(select, 2);
 		msg->uid = (unsigned long) sqlite3_column_int64(select, 4);
 		msg->gid = (unsigned long) sqlite3_column_int64(select, 5);
